/*
   ** Copyright (C) 2000,2001 James McKenzie
   **
   ** Filename:    samsung.c
   ** Author:      James McKenzie
   ** Date:        12 January 1999, Mar 22 2001
   ** Version:     1.0
   ** Platform:    Linux
   ** WWW ref:     http://bullard.esc.cam.ac.uk/~james/digimax.html
   ** Licence:     GNU Public Licence
   **
   ** This program is free software; you can redistribute it and/or modify
   ** it under the terms of the GNU General Public License as published by
   ** the Free Software Foundation; version 2 dated June, 1991.
   **
   ** This program is distributed in the hope that it will be useful,
   ** but WITHOUT ANY WARRANTY; without even the implied warranty of
   ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   ** GNU General Public License for more details.
   **
   ** You should have received a copy of the GNU General Public License
   ** along with this program;  if not, write to the Free Software
   ** Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
   **
   ** For more infomation see http://www.gnu.org/copyleft/gpl.html
   **
   ** Description:
   **  This program downloads images from a Samsung Digimax 800k
   **
   ** Todo list:
   **   - fix crc code
   **
 */

#include<stdio.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<termios.h>
#include<sys/ioctl.h>
#include<setjmp.h>
#include<sys/time.h>
#include<signal.h>
#include <string.h>
#include "../src/gphoto.h"
#include "../src/util.h"


/* print all data transmitted and received via the serial port */
/* #define DEBUG */

typedef unsigned char byte;

/* Camera responses */
const unsigned char SDSC_EOF = 0x04;
const unsigned char SDSC_OK = 0x06;
const unsigned char SDSC_IEXIST = 0x49;

/* Commands */
const unsigned char SDSC_BINARY = 0x06;
const unsigned char SDSC_RETRANSMIT = 0x15;
const unsigned char SDSC_GIMME = 0x42;
const unsigned char SDSC_START = 0x43;
const unsigned char SDSC_DELETE = 0x44;
const unsigned char SDSC_GOODBYE = 0x45;
const unsigned char SDSC_NEXT = 0x53;
const unsigned char SDSC_HELLO = 0x57;

/* number of image data bytes transfered in each packet */
const int SDSC_BLOCKSIZE = 0x405; /* 3 bytes header, 1024 bytes data, 2 crc */
const int SDSC_INFOSIZE = 0x85; /* 3 bytes header, 128 bytes data , 2 crc */
const int SDSC_TIMEOUT = 500000; /* usec */
const int SDSC_RETRIES = 3;


/* Protocols supported, state of camera*/
enum
{
  PROTO_UNK = 0,                /*Unknown */
  PROTO_1 = 1,                  /*Proto 1 */
  PROTO_2 = 2,                  /*Proto 2 */
  PROTO_IDLE = 3,               /*Idle (for mode) */
  PROTO_2I = 4                  /*Proto 2 Idle (Still can't use proto 1 commands) */
};

int fd;
int camera_cap = PROTO_UNK;     /*Camera capabilities */
int camera_mode = PROTO_IDLE;   /*Current state */
int camera_npics = -1;          /*Number of pictures */

/* Thumbnail storage*/
int numnails = -1;

struct thumb_struct
{
  byte *image;
  int size;
  byte *name;
}
 *thumbnails = NULL;


#ifdef STANDALONE
void
error_dialog (char *Error)
{
  fprintf (stderr, "ERROR_DIAGLOG: %s\n", Error);
}

void
update_status (char *Error)
{
  fprintf (stderr, "UPDATE_PROGRESS: %s\n", Error);
}

void
update_progress (float f)
{
  fprintf (stderr, "UPDATE_PROGRESS: %f\n", f);
}


char serial_port[20] = "/dev/ttyS0";
#endif
/****************** Low level IO *********************/

#define DAZED error_dialog("Confused, quit gphoto and power cycle camera")

static int
sendcommand (byte cmd)
{
  int i;

#ifdef DEBUG
  fprintf (stderr, "SDSC> %02x\n", cmd);
#endif
  return write (fd, &cmd, 1);
}


#ifdef DEBUG
static void
dump_bytes (byte * buf, int len)
{
  int i;
  for (i = 0; i < len; ++i)
    {
      if ((i % 16) == 0)
        {
          if (i)
            fprintf (stderr, "\n");
          fprintf (stderr, "SDSC<");
        }
      fprintf (stderr, " %02x", buf[i]);
    }
  fprintf (stderr, "\n");
}
#endif

static int
waitforinput ()
{
  fd_set rfds;
  struct timeval tv;

  tv.tv_sec = 0;
  tv.tv_usec = SDSC_TIMEOUT;

  FD_ZERO (&rfds);
  FD_SET (fd, &rfds);

  if (!select (fd + 1, &rfds, NULL, NULL, &tv))
    return 0;

  if (!FD_ISSET (fd, &rfds))
    return 0;

  return 1;
}

static int
checkcrc (byte * buf, int len)
{
/* FIXME: */
  return 1;
}

static int
receive (byte * buf, int length, int expected, int retry)
{
  int tries = 0;
  int n, total, i;
  byte *ptr;


#ifdef DEBUG
  fprintf (stderr, "receive(%p,%d,%d)\n", buf, length, expected);
#endif
  while (1)
    {
      ptr = buf;
      i = length;
      total = 0;

      while (waitforinput (fd))
        {

          n = read (fd, ptr, i);

          if (n > 0)
            {
              ptr += n;
              i -= n;
              total += n;
            }

          if (total == expected)
            {

#ifdef DEBUG
              fprintf (stderr, "Read %d bytes as expected:\n", total);
              dump_bytes (buf, total);
#endif

              if (checkcrc (buf, total))
                {
                  return total;
                }
              else
                {;
#ifdef DEBUG
                  fprintf (stderr, "CRC bad - waiting for anything else\n");
#endif
                }


            }
        }

#ifdef DEBUG
      fprintf (stderr, "Read %d bytes where %d bytes expected\n", total,
               expected);
      dump_bytes (buf, total);
#endif

      /*Ok didn't match expected - two posibities */
      /*1st we missed packet/crcerror */
      /*2nd this is the last packet */

      tries++;


      if ((total == 1) && (tries == 2) && (buf[0] == SDSC_EOF))
        {
#ifdef DEBUG
          fprintf (stderr, "Detected EOF condition\n");
#endif
          return total;
        }

      if ((tries == SDSC_RETRIES) || (!retry))
        {
#ifdef DEBUG
          fprintf (stderr, "Giving up\n");
#endif
          if (!retry)
            DAZED;
          return 0;
        }

#ifdef DEBUG
      fprintf (stderr, "Retrying\n");
#endif
      sendcommand (SDSC_RETRANSMIT);


    }

}


static int
init_comms ()
{
  struct termios tios;
  unsigned char buf[0x100];
  int length;

  if (tcgetattr (fd, &tios) < 0)
    error_dialog ("Error getting communication parameters");

  tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB | CLOCAL | CRTSCTS);
  tios.c_cflag |= CLOCAL;
  tios.c_cflag |= CS8;

  tios.c_cflag |= CREAD | HUPCL;
  tios.c_iflag = IGNBRK | IGNPAR;

  tios.c_oflag = 0;
  tios.c_lflag = 0;
  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 0;


  cfsetospeed (&tios, B115200);
  cfsetispeed (&tios, B115200);

  if (tcsetattr (fd, TCSAFLUSH, &tios) < 0)
    error_dialog ("Error setting communication parameters");

  return 0;
}


/*********************** Mid level IO *************************/

static int
isnullinfo (byte * buf)
{
  int n = 128;
  buf += 3;

  while (n--)
    {
      if (*(buf++))
        return 0;
    }

  return 1;
}

static int
proto1_makeidle ()
{
  byte buf[2048];
  int len;


  while (1)
    {
      sendcommand (SDSC_NEXT);
      sendcommand (SDSC_START); /*Read out a header */
      len = receive (buf, sizeof (buf), SDSC_INFOSIZE, 1);

      if (len != SDSC_INFOSIZE)
        {
          camera_mode = PROTO_UNK;
          DAZED;

          return 0;
        }

      if (isnullinfo (buf))
        break;

    }

  sendcommand (SDSC_NEXT);

  return 1;

}


static int
proto2_makeidle ()
{
  byte buf[2048];
  int len;

  sendcommand (SDSC_GOODBYE);
  len = receive (buf, sizeof (buf), 1, 0);
  if ((len == 1) && (buf[0] == SDSC_OK))
    {
      camera_mode = PROTO_2I;
      return 1;
    }

  camera_mode = PROTO_UNK;
  return 0;
}



static int
makeidle (void)
{
  switch (camera_mode)
    {
    case PROTO_IDLE:
    case PROTO_2I:
      return 1;
    case PROTO_1:
      return proto1_makeidle ();
    case PROTO_2:
      return proto2_makeidle ();
    default:
      return 0;

    }
}


void
free_thumbnails (void)
{
  int i;

#ifdef DEBUG
  fprintf (stderr, "FTH: %p %d\n", thumbnails, numnails);
#endif

  if (!thumbnails)
    return;


  for (i = 0; i < numnails; ++i)
    {
#ifdef DEBUG
      fprintf (stderr, "FTH: %p %p\n", thumbnails[i].image,
               thumbnails[i].name);
#endif
      if (thumbnails[i].image)
        free (thumbnails[i].image);
      if (thumbnails[i].name)
        free (thumbnails[i].name);
    }
  free (thumbnails);
  numnails = -1;
  thumbnails = NULL;

  return;
}

static int
numpics (void)
{
  byte buf[2048];
  int len;

  if (camera_npics >= 0)
    return camera_npics;
  if (!makeidle ())
    return -1;

  camera_npics = 0;

  while (1)
    {
      sendcommand (SDSC_NEXT);
      sendcommand (SDSC_START); /*Read out a header */
      len = receive (buf, sizeof (buf), SDSC_INFOSIZE, 1);

      if (len != SDSC_INFOSIZE)
        {
          camera_mode = PROTO_UNK;

          return -1;
        }

      if (isnullinfo (buf))
        break;

      camera_npics++;
    }


  sendcommand (SDSC_NEXT);

  return camera_npics;

}


static int
proto2_enter ()
{
  byte buf[2048];
  char str[1024];
  int i, n, tries, len;
  int blocks, size, left;
  int th;
  byte *ptr;

  if (camera_mode == PROTO_2)
    return 1;
  if (camera_mode == PROTO_2I)
    return 1;
  if (camera_cap == PROTO_1)
    return 0;

/*need this for later*/
  numpics ();

/* Quit proto 1*/
  if ((camera_mode == PROTO_1) && (!makeidle ()))
    return 0;

  camera_mode = PROTO_UNK;

  tries = 0;
  while (1)
    {
      sendcommand (SDSC_HELLO);
      len = receive (buf, sizeof (buf), 1, 0);
      if ((len == 1) && (buf[0] == SDSC_IEXIST))
        break;
      sleep (1);
      tries++;
      if (tries == SDSC_RETRIES)
        {
#ifdef DEBUG
          fprintf (stderr, "FAILED: setting camera cap %d\n", camera_cap);
#endif

          error_dialog ("You have the early version of Digimax 800k\n"
                        "It doesn't support getting thumbnails");

          switch (camera_cap)
            {
            case PROTO_UNK:
            case PROTO_1:
              camera_cap = PROTO_1;
              camera_mode = PROTO_IDLE;
              return 0;
            default:
              camera_mode = PROTO_UNK;
              error_dialog ("Confused, quit and power cycle the camera");
              return 0;

            }
        }
    }

  camera_cap = PROTO_2;
  free_thumbnails ();

  numnails = numpics ();
  thumbnails =
    (struct thumb_struct *) malloc ((sizeof (struct thumb_struct)) *
                                    (numnails + 2));
  bzero (thumbnails, (sizeof (struct thumb_struct)) * (numnails + 2));

  th = 1;

  while (1)
    {

//FIXME: make entry
      sendcommand (SDSC_START);
      len = receive (buf, sizeof (buf), SDSC_INFOSIZE, 1);
#ifdef DEBUG
      {
        byte *ptr = buf + 15;
        fprintf (stderr, "INFO: file %s size %s\n", buf + 3, ptr);
      }
#endif
      if (isnullinfo (buf))
        break;

      buf[SDSC_INFOSIZE] = 0;

      thumbnails[th].size = size = atoi (&buf[15]);
      thumbnails[th].name = strdup (&buf[3]);


      sprintf (str, "Thumbnail %d is %d bytes\n", th, size);
      update_status (str);

      blocks = (size + 0x3ff) / 0x400;
      blocks += 2;

      left = size = 0x400 * blocks;
      i = 0;

      ptr = thumbnails[th].image = (byte *) malloc (size);

      if (!ptr)
        {
          free_thumbnails ();
          error_dialog ("Failed to malloc image data.");
          return 0;
        }

      update_progress (0.0);


      sendcommand (SDSC_BINARY);
      sendcommand (SDSC_START);


      while (left >= 0x400)
        {
          len = receive (buf, sizeof (buf), SDSC_BLOCKSIZE, 1);

          if (!len)
            {
              free_thumbnails ();

#ifdef DEBUG
              fprintf (stderr, "Get thumbnail failed \n");
#endif

              return 0;
            }

          sendcommand (SDSC_BINARY); /*Next block */

          if ((len == 1) && (buf[0] == SDSC_EOF))
            break;

          bcopy (buf + 3, ptr, 0x400);

          ptr += 0x400;
          left -= 0x400;
          i += 0x400;

          update_progress ((float) i / (float) size);

        }
      th++;
    }

  sendcommand (SDSC_BINARY);

  camera_mode = PROTO_2;

  return 1;
}

int
seek_and_start (picnum)
{
  int len;
  byte buf[2048];

  switch (camera_mode)
    {
    case PROTO_1:
      makeidle ();
      /* Fall thru */
    case PROTO_IDLE:
      while (--picnum)
        {
          sendcommand (SDSC_START);
          len = receive (buf, sizeof (buf), SDSC_INFOSIZE, 1);
          if (len != SDSC_INFOSIZE)
            {
              camera_mode = PROTO_UNK;
              DAZED;
              return 0;
            }
          sendcommand (SDSC_NEXT);
        }
      sendcommand (SDSC_START);
      camera_mode = PROTO_1;
      return 1;
    case PROTO_2:
    case PROTO_2I:
      sendcommand (SDSC_GIMME);
      sprintf (buf, "PIC%04d.JPG", picnum);
      for (len = 0; buf[len]; len++)
        sendcommand (buf[len]);
      sendcommand (0);
      return 1;
    default:
      return 0;
    }
}


/******************** Top level if ****************************/


/*
 * Initialize comms to the camera...
 */
static int
SDSC_Initialise (void)
{

  fd = open (serial_port, O_RDWR);

#ifdef DEBUG
  fprintf (stderr, "RESET ALL\n");
#endif

  camera_cap = PROTO_UNK;
  camera_mode = PROTO_IDLE;
  camera_npics = -1;

  free_thumbnails ();


  if (fd < 0)
    {
      error_dialog ("Couldn't open camera device");
      return 1;
    }

  if (init_comms (fd))
    {
      error_dialog ("Couldn't init comm port");
      return 1;
    }
  return 0;

}

/*
   **
 */
static int
SDSC_disconnect ()
{
#ifdef DEBUG
  fprintf (stderr, "Do disconnect\n");
#endif


  makeidle ();

  free_thumbnails ();

#ifdef DEBUG
  fprintf (stderr, "Done disconnect\n");
#endif

  return 0;
}

/*
   ** returns 0 for error or the number of pictures in the camera
 */
static int
SDSC_QueryPics ()
{
  if (camera_mode == PROTO_UNK)
    return 0;
  return numpics ();
}


/*
   ** fill out the rest of the image structure, given "number" already
 */
static struct Image *
SDSC_ReadImage (int picnum, int thumbnail)
{
  struct Image *image = (struct Image *) malloc (sizeof (struct Image));
  int len, i, left, size, blocks;
  byte buf[2048], *ptr;
  char str[1024];

#ifdef DEBUG
  fprintf (stderr, "READIMAGE %d %d %d\n", picnum, thumbnail, camera_cap);
#endif

  if (camera_mode == PROTO_UNK)
    return NULL;

  if (thumbnail)
    {
      if (!thumbnails)
        proto2_enter ();
      if ((!thumbnails) || (picnum < 1) || (picnum > numnails))
        {
          free (image);
          return NULL;
        }

      image->image = malloc (thumbnails[picnum].size);
      if (!image->image)
        {
          free (image);
          return NULL;
        }


      bcopy (thumbnails[picnum].image, image->image, thumbnails[picnum].size);
      image->image_size = thumbnails[picnum].size;
      strcpy (image->image_type, "jpg");
      image->image_info_size = 0;

      return image;

    }


  if (!seek_and_start (picnum))
    {
      free (image);
      return NULL;
    }

  len = receive (buf, sizeof (buf), SDSC_INFOSIZE, 1);

  if (len != SDSC_INFOSIZE)
    {
      free (image);
      camera_mode = PROTO_UNK;
      DAZED;
      return NULL;
    }

  buf[SDSC_INFOSIZE] = 0;

  image->image_size = atoi (&buf[15]);

  sprintf (str, "Length of image %d is %d bytes\n",
           picnum, image->image_size);
  update_status (str);

  blocks = (image->image_size + 0x3ff) / 0x400;
  blocks += 2;

  left = size = 0x400 * blocks;
  i = 0;

  ptr = image->image = (unsigned char *) malloc (size);

  if (!image->image)
    {
      error_dialog ("Failed to malloc image data.");
      free (image);
      return 0;
    }

  update_progress (0.0);

  sendcommand (SDSC_BINARY);    /*Put into image mode */
  sendcommand (SDSC_START);     /*Start reading */

  while (left >= 0x400)
    {
      len = receive (buf, sizeof (buf), SDSC_BLOCKSIZE, 1);

      if (!len)
        {
          free (image->image);
          free (image);
#ifdef DEBUG
          fprintf (stderr, "ReadImage failed \n");
#endif
          return 0;
        }

      if ((len == 1) && (buf[0] == 0x4))
        break;

      bcopy (buf + 3, ptr, 0x400);

      ptr += 0x400;
      left -= 0x400;
      i += 0x400;

      update_progress ((float) i / (float) size);

      sendcommand (SDSC_BINARY); /*Next block */
    }

  sendcommand (SDSC_BINARY);    /*Next block */

  strcpy (image->image_type, "jpg");
  image->image_info_size = 0;
#ifdef DEBUG
  fprintf (stderr, "Done ReadImage \n");
#endif

  return image;
}

/*
   ** delete an image from the camera
   ** The image number is one based (1 is the first image).
 */
static int
SDSC_DeleteImage (int image_no)
{
  return 0;
}

static char *
SDSC_Summary (void)
{
  return "Not Supported";
}

static char *
SDSC_Description (void)
{
  return "gPhoto Samsung digimax 800k driver\n"
    "(C) James Mckenzie 2000\n" "mailto:james@fishsoup.dhs.org\n";
}

static struct Image *
SDSC_Preview (void)
{
  return 0;
}

static int
SDSC_TakePicture ()
{
  return 0;
}

static int
SDSC_Configure ()
{
  return 0;
}

struct _Camera samsung800k = {
  SDSC_Initialise,
  SDSC_ReadImage,
  SDSC_Preview,
  SDSC_DeleteImage,
  SDSC_TakePicture,
  SDSC_QueryPics,
  SDSC_Configure,
  SDSC_Summary,
  SDSC_Description,
};



#ifdef STANDALONE
main ()
{

  SDSC_Initialise ();

  printf ("Pictures = %d\n", SDSC_QueryPics ());

  fprintf (stderr, "FISH\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");

  SDSC_ReadImage (2, 3);



}
#endif
